==> tuxurf-0.01.tar.gz: HTW/main.c <==
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ghttp.h>

#include "parser.h"


#define MAIN_FONT "-adobe-helvetica-%s-r-normal--%d-*-*-*-*-*-iso8859-1"
#define FIXED_FONT "-adobe-courier-%s-r-normal--%d-*-*-*-m-*-iso8859-1"


static char buf[16384];
static char *pagebuf = buf;
static ghttp_request *http_request;
static char *base_url;
static char *base_host;
static char *base_path;

static int is_full_url(char *p)
{
	char *colon = strchr(p,':');
	char *slash = strchr(p, '/');
	if(colon && (slash==NULL || colon < slash))
		return 1;
	return 0;
}

static void set_url_pointers(void)
{
	char *p=strchr(base_url,':');
	base_host = ++p;
	while(*p=='/')
		p++;
	p=strchr(p, '/');
	if(p)
		base_path = p;
	else
		base_path = base_url+strlen(base_url);
}

void set_base_url(char *p)
{
	char *segment;
	
	if(base_url)
		set_url_pointers();
	
	if(is_full_url(p))
	{
		if(base_url)
			free(base_url);
		base_url = strdup(p);
		return;
	}
	if(*p=='/')
	{
		/* Host relative */
		char *buf=malloc(strlen(base_url)-strlen(base_path)+strlen(p)+2);
		*base_path=0; /* The '/' */
		sprintf(buf, "%s%s", base_url, p);
		if(base_url)
			free(base_url);
		base_url = buf;
		return;
	}
	segment = strrchr(base_path, '/');
	if(segment)
		*segment = 0;
	{
		char *buf=malloc(strlen(base_url)+strlen(p)+2);
		sprintf(buf, "%s/%s", base_url, p);
		if(base_url)
			free(base_url);
		base_url = buf;
	}
}	

static GtkWidget *app, *draw, *html, *scroller;
static GtkObject *scrollpos;
static GdkPixmap *pixmap = NULL;
static int pixmap_dirty;
static GdkFont *html_font[3][2];
#define FONT_SMALL	0
#define FONT_MEDIUM	1
#define FONT_BIG	2
static GdkColor blue;
static GdkGC *bluecol;
static struct render_state gtk_rt;

static int scrolltop = 0;
static int scrollbottom = 0;
static int pixwidth = 0;
static int pixheight = 0;
static int spacesize;
static int spacesizes[3];
static int pos_y;
int lastnl = 1;
static int pos_x;
static int rowsize;
static int rowsizes[3];
static char wordbuf[64];
static int wordp;
static char titlebuf[64];


static int fontsize;
static int fontbold;

static int wordsize(void)
{
	return gdk_text_width(html_font[fontsize][fontbold], wordbuf, wordp);
}

static void draw_text(char *p)
{
	GdkRectangle update_rect;
	GdkGC *gc;
	GdkFont *font;
	
	if(bluecol==NULL)
	{
		bluecol=gdk_gc_new(html->window);
		gdk_gc_copy(bluecol, html->style->black_gc);
		gdk_gc_set_foreground(bluecol, &blue);
	}
	
	update_rect.x = pos_x;
	update_rect.y = pos_y - scrolltop;
	update_rect.height = rowsize;
	update_rect.width = wordsize();
	
	gc = html->style->black_gc;
	font = html_font[fontsize][fontbold];
	
	if(gtk_rt.rfont & RF_LINK)
	{
		gc = bluecol;
	}
	
	if(pixmap && pos_y+rowsize >= scrolltop)
	{
		if((gtk_rt.rfont & RF_LINK) && link_top())
			link_add_box(link_top(), pos_x, pos_y-scrolltop-rowsize, update_rect.width, rowsize);

		gdk_draw_string(pixmap, font,
				gc,
				pos_x, pos_y-scrolltop,p);
//		printf("Wrote %s at %d,%d\n", p, pos_x, pos_y-scrolltop);
	}
	gtk_widget_draw(html, &update_rect);
}

static void select_font(void)
{
	if(render->rfont&RF_BOLD)
		fontbold = 1;
	else
		fontbold = 0;
		
	if(render->rsize == 1 || render->rsize == 2)
		fontsize = FONT_BIG;
	else if(render->rsize == 3 || render->rsize == 4)
		fontsize = FONT_MEDIUM;
	else if(render->rsize == 5 || render->rsize == 6)
	{
		fontsize = FONT_SMALL;
		fontbold = 1;
	}
	else
		fontsize = FONT_SMALL;

	if(rowsize < rowsizes[fontsize])		
		rowsize = rowsizes[fontsize];
	spacesize = spacesizes[fontsize];
}

static void flush_word(int s)
{
	static int lastp = 1;
	select_font();
	
	if(wordp || lastp ==0)
	{
		lastp = s;
		wordbuf[wordp]=0;
		draw_text(wordbuf);
		pos_x+=wordsize();
		if(s)
			pos_x+=spacesize;
		wordp=0;
		lastnl=0;
	}
}

void decode_entities(char *p)
{
	char *o = p;
	
	while(*p)
	{
		if(*p!='&')
			*o++=*p++;
		else
		{
			char *ee=strchr(++p,';');
			if(ee==NULL)
			{
				memmove(o,p, strlen(p));
				break;
			}
			*ee = 0;
			
			if(strcasecmp(p, "amp")==0)
				*o++='&';
			else if(strcasecmp(p, "copy")==0)
				*o++='\xA9';
			else if(strcasecmp(p, "gt")==0)
				*o++='>';
			else if(strcasecmp(p, "lt")==0)
				*o++='<';
			else if(strcasecmp(p, "quot")==0)
				*o++='"';
			else if(strcasecmp(p, "reg")==0)
				*o++='\xAE';
			else if(strcasecmp(p, "nbsp")==0)
				*o++='\xA0';
			else if(*p=='#')
			{
				int n = atoi(p+1);
				if((n<32 || n> 127) && n!=160)
					*o++=' ';
				else
					*o++=(char)n;
			}
			else
				*o++='?';
			p=ee+1;
		}
	}
	*o=0;
}


void render_text(char *p)
{
	decode_entities(p);
	
	if(render->rhead)
	{
		if(render->rlayout&RL_TITLE)
		{
			strncat(titlebuf, p, 63);
			titlebuf[63]=0;
			gtk_window_set_title(GTK_WINDOW(app), titlebuf);
			return;
		}
	}
	
	if(render->rfont&RF_HIDE)
		return;
		
	while(*p)
	{
		if(isspace(*p))
			flush_word(1);
		else 
		{
			if(wordp == 63)
				flush_word(0);
			else if(pos_x+wordsize()>pixwidth)
			{
				pos_y += rowsize;
				pos_x=0;
				lastnl=1;
				rowsize=rowsizes[0];
			}
			wordbuf[wordp++]=*p;
		}
		p++;
	}
	flush_word(0);
}


void renderer_reset(void)
{
	int boldspace, boldrow;
	int i;
	
	lastnl = 1;
	wordp = 0;
	pos_x = 0;
	
	for(i=0;i<3;i++)
	{
		spacesizes[i] = gdk_char_width(html_font[i][0], ' ');
		rowsizes[i] = gdk_char_height(html_font[i][0], 'X');

		boldspace = gdk_char_width(html_font[i][1], ' ');
		boldrow = gdk_char_height(html_font[i][1], 'X');
	
		/*
		 *	Allow for both fonts
		 */

		if(boldrow > rowsizes[i])
			rowsizes[i] = boldrow;
		if(boldspace > spacesizes[i])
			spacesizes[i] = boldspace;
	
		rowsizes[i]+=3;
	}		

	pos_y = rowsizes[0];

	*titlebuf=0;
}

void renderer_begin(struct token_stack *t)
{
	if(pixmap)
		gdk_draw_rectangle(pixmap,
			html->style->white_gc,
			TRUE,
			0, 0,
			html->allocation.width,
			html->allocation.height);
	link_free();
}


void render_newline(void)
{
	flush_word(0);
	pos_y += rowsize;
	lastnl=1;
	pos_x=0;
	rowsize=rowsizes[0];
}

void render_newline_p(void)
{
	flush_word(0);
	if(lastnl==0)
		render_newline();
}

void renderer_complete(struct token_stack *t)
{
	render_newline_p();
	GTK_ADJUSTMENT(scrollpos)->upper = (gfloat) pos_y;
	GTK_ADJUSTMENT(scrollpos)->step_increment = (gfloat) rowsize;
	GTK_ADJUSTMENT(scrollpos)->page_size = (gfloat)pixheight;
	if(pixheight > 2*rowsize)
		GTK_ADJUSTMENT(scrollpos)->page_increment = (gfloat)(pixheight-rowsize);
	else
		GTK_ADJUSTMENT(scrollpos)->page_increment = (gfloat)pixheight;
	gtk_adjustment_changed(GTK_ADJUSTMENT(scrollpos));
}



static gint resize_html(GtkWidget *w, GdkEventConfigure *event)
{
	static int n = 0;
	static int oldh= -1, oldw = -1;

	if(oldh != w->allocation.height || oldw != w->allocation.width)
	{
		if(!pixmap_dirty)
			pixmap_dirty = 1;
		
		if(pixmap)
			gdk_pixmap_unref(pixmap);
		pixmap = gdk_pixmap_new(w->window,
					w->allocation.width,
					w->allocation.height,
					-1);
		gdk_draw_rectangle(pixmap,
			w->style->white_gc,
			TRUE,
			0, 0,
			w->allocation.width,
			w->allocation.height);
		scrollbottom = scrolltop + w->allocation.height;
		pixwidth = w->allocation.width - 16;
		pixheight = w->allocation.height;
#if 0	
		render_reset(&gtk_rt);
		parser(pagebuf);
		parse_complete();
#endif
		oldh = w->allocation.height;
		oldw = w->allocation.width;
	}
	return TRUE;
}

static gint draw_html(GtkWidget *w, GdkEventExpose *event)
{
	if(pixmap_dirty == 1)
	{
		pixmap_dirty = 2;
		render_reset(&gtk_rt);
		parser(pagebuf);
		parse_complete();
		pixmap_dirty = 0;
	}
	
	if(pixmap)
		gdk_draw_pixmap(w->window,
			w->style->fg_gc[GTK_WIDGET_STATE(w)],
			pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
	return FALSE;
}

static void mark_html_dirty(void)
{
	GdkRectangle update_rect;
	
	if(!pixmap_dirty)
	{
		pixmap_dirty = 1;

		update_rect.x = 0;
		update_rect.y = 0;
		update_rect.height = pixheight;
		update_rect.width = pixwidth + 16;

		gtk_widget_draw(html, &update_rect);
	}
}

static gint scroll_html(GtkWidget *w, gpointer unused)
{
	static int lastscroll;
	
	scrolltop = (int)GTK_ADJUSTMENT(scrollpos)->value;
	scrollbottom = scrolltop + pixheight;
	
	if(scrolltop != lastscroll)
	{
		lastscroll = scrolltop;
		mark_html_dirty();
	}
}

static gint click_html(GtkWidget *w, GdkEventButton *b)
{
	if(b->button == 1)
	{
		struct link_url *l = link_hit(b->x, b->y);
		if(l!=NULL)
		{
			printf("Go to %s\n", l->data);
			set_base_url(l->data);
			printf("Going to %s\n", base_url);
			/* Hack loader */
			if(http_request)
			{
				*buf=0;
				pagebuf=buf;
				ghttp_request_destroy(http_request);	
				link_free();
			}
			gtk_adjustment_set_value(GTK_ADJUSTMENT(scrollpos),0.0);
			http_request = ghttp_request_new();
			ghttp_set_uri(http_request, base_url);
			ghttp_prepare(http_request);
			ghttp_process(http_request);
			pagebuf = ghttp_get_body(http_request);
			/* HACK!! */
			pagebuf[ghttp_get_body_len(http_request)-1]=0;
			mark_html_dirty();
		}
	}
}

int main(int argc, char *argv[])
{
	int i;
		/* FIXME */
	int fd=open(argv[1]?argv[1]:"foo.html", O_RDONLY);
	int l = read(fd, buf, 16384);
	close(fd);
	buf[l]=0;	

	
	gtk_init(&argc, &argv);
	app = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect(GTK_OBJECT(app), "destroy", 
		GTK_SIGNAL_FUNC(gtk_exit), NULL);
	gtk_window_set_title(GTK_WINDOW(app), "GWeb");
	gtk_container_set_border_width(GTK_CONTAINER(app), 0);
	
	draw = gtk_hbox_new(FALSE,0);
	gtk_container_set_border_width(GTK_CONTAINER(draw), 2);

	gtk_container_add(GTK_CONTAINER(app), draw);

	for(i=0;i<3;i++)
	{
		char fontbuf[128];
		sprintf(fontbuf, MAIN_FONT, "medium", 12+2*i);
		html_font[i][0] = gdk_font_load(fontbuf);
		if(html_font[i][0] == NULL)
		{
			fprintf(stderr, "Font not found!\n");
			exit(1);
		}
		gdk_font_ref(html_font[i][0]);
		sprintf(fontbuf, MAIN_FONT, "bold", 12+2*i);
		html_font[i][1] = gdk_font_load(fontbuf);
		if(html_font[i][1] == NULL)
		{
			fprintf(stderr, "Font not found!\n");
			exit(1);
		}
		gdk_font_ref(html_font[i][1]);
	}

	gdk_color_parse("#0000FF", &blue);
	gdk_colormap_alloc_color(gdk_colormap_get_system(), &blue, FALSE, TRUE);

	html = gtk_drawing_area_new();
	gtk_box_pack_start(GTK_BOX(draw), html, TRUE, TRUE, 0);
	
	scrollpos = gtk_adjustment_new(0.0,0.0,0.0,0.0,0.0,0.0);
	scroller = gtk_vscrollbar_new(GTK_ADJUSTMENT(scrollpos));
	gtk_box_pack_start(GTK_BOX(draw), scroller, FALSE, FALSE, 0);
	
	gtk_signal_connect(GTK_OBJECT(html), "expose_event", (GtkSignalFunc)draw_html, NULL);
	gtk_signal_connect(GTK_OBJECT(html), "configure_event", (GtkSignalFunc)resize_html, NULL);
	gtk_signal_connect(GTK_OBJECT(html), "button_press_event", (GtkSignalFunc)click_html, NULL);

	gtk_widget_set_events(html, GDK_EXPOSURE_MASK|GDK_BUTTON_PRESS_MASK);
		
	gtk_signal_connect(GTK_OBJECT(scrollpos), "value-changed", (GtkSignalFunc)scroll_html, NULL);
	
	gtk_widget_show(html);
	gtk_widget_show(scroller);
	gtk_widget_show(draw);
	gtk_widget_show(app);
	gtk_main();
	return 0;
}

==> tuxurf-0.02.tar.gz: HTW/main.c <==
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <ghttp.h>

#include "parser.h"


#define MAIN_FONT "-adobe-helvetica-%s-r-normal--%d-*-*-*-*-*-iso8859-1"
#define FIXED_FONT "-adobe-courier-%s-r-normal--%d-*-*-*-m-*-iso8859-1"


char buf[16384];
char *pagebuf = buf;

static GtkWidget *app, *draw, *html, *scroller;
static GtkObject *scrollpos;
static GdkPixmap *pixmap = NULL;
static int pixmap_dirty;
static GdkFont *html_font[3][2];
#define FONT_SMALL	0
#define FONT_MEDIUM	1
#define FONT_BIG	2
static GdkColor blue;
static GdkGC *bluecol;
static struct render_state gtk_rt;

static int scrolltop = 0;
static int scrollbottom = 0;
static int pixwidth = 0;
static int pixheight = 0;
static int spacesize;
static int spacesizes[3];
static int pos_y;
int lastnl = 1;
static int pos_x;
static int rowsize;
static int rowsizes[3];
static char wordbuf[64];
static int wordp;
static char titlebuf[64];


static int fontsize;
static int fontbold;

static int wordsize(void)
{
	return gdk_text_width(html_font[fontsize][fontbold], wordbuf, wordp);
}

static void draw_text(char *p)
{
	GdkRectangle update_rect;
	GdkGC *gc;
	GdkFont *font;
	
	if(bluecol==NULL)
	{
		bluecol=gdk_gc_new(html->window);
		gdk_gc_copy(bluecol, html->style->black_gc);
		gdk_gc_set_foreground(bluecol, &blue);
	}
	
	update_rect.x = pos_x;
	update_rect.y = pos_y - scrolltop;
	update_rect.height = rowsize;
	update_rect.width = wordsize();
	
	gc = html->style->black_gc;
	font = html_font[fontsize][fontbold];
	
	if(gtk_rt.rfont & RF_LINK)
	{
		gc = bluecol;
	}
	
	if(pixmap && pos_y+rowsize >= scrolltop)
	{
		if((gtk_rt.rfont & RF_LINK) && link_top())
			link_add_box(link_top(), pos_x, pos_y-scrolltop-rowsize, update_rect.width, rowsize);

		gdk_draw_string(pixmap, font,
				gc,
				pos_x, pos_y-scrolltop,p);
//		printf("Wrote %s at %d,%d\n", p, pos_x, pos_y-scrolltop);
	}
	if(!pixmap_dirty)
		gtk_widget_draw(html, &update_rect);
}

static void select_font(void)
{
	if(render->rfont&RF_BOLD)
		fontbold = 1;
	else
		fontbold = 0;
		
	if(render->rsize == 1 || render->rsize == 2)
		fontsize = FONT_BIG;
	else if(render->rsize == 3 || render->rsize == 4)
		fontsize = FONT_MEDIUM;
	else if(render->rsize == 5 || render->rsize == 6)
	{
		fontsize = FONT_SMALL;
		fontbold = 1;
	}
	else
		fontsize = FONT_SMALL;

	if(rowsize < rowsizes[fontsize])		
		rowsize = rowsizes[fontsize];
	spacesize = spacesizes[fontsize];
}

static void flush_word(int s)
{
	static int lastp = 1;
	select_font();
	
	if(wordp || lastp ==0)
	{
		lastp = s;
		wordbuf[wordp]=0;
		draw_text(wordbuf);
		pos_x+=wordsize();
		if(s)
			pos_x+=spacesize;
		wordp=0;
		lastnl=0;
	}
}

void decode_entities(char *p)
{
	char *o = p;
	
	while(*p)
	{
		if(*p!='&')
			*o++=*p++;
		else
		{
			char *ee=strchr(++p,';');
			if(ee==NULL)
			{
				memmove(o,p, strlen(p));
				break;
			}
			*ee = 0;
			
			if(strcasecmp(p, "amp")==0)
				*o++='&';
			else if(strcasecmp(p, "copy")==0)
				*o++='\xA9';
			else if(strcasecmp(p, "gt")==0)
				*o++='>';
			else if(strcasecmp(p, "lt")==0)
				*o++='<';
			else if(strcasecmp(p, "quot")==0)
				*o++='"';
			else if(strcasecmp(p, "reg")==0)
				*o++='\xAE';
			else if(strcasecmp(p, "nbsp")==0)
				*o++='\xA0';
			else if(*p=='#')
			{
				int n = atoi(p+1);
				if((n<32 || n> 127) && n!=160)
					*o++=' ';
				else
					*o++=(char)n;
			}
			else
				*o++='?';
			p=ee+1;
		}
	}
	*o=0;
}


void render_text(char *p)
{
	decode_entities(p);
	
	if(render->rhead)
	{
		if(render->rlayout&RL_TITLE)
		{
			strncat(titlebuf, p, 63);
			titlebuf[63]=0;
			gtk_window_set_title(GTK_WINDOW(app), titlebuf);
			return;
		}
	}
	
	if(render->rfont&RF_HIDE)
		return;
		
	while(*p)
	{
		if(isspace(*p))
			flush_word(1);
		else 
		{
			if(wordp == 63)
				flush_word(0);
			else if(pos_x+wordsize()>pixwidth)
			{
				if(pos_x == 0)
				{
					flush_word(0);
				}
				else
				{
					pos_y += rowsize;
					pos_x=0;
					lastnl=1;
					rowsize=rowsizes[0];
				}
			}
			wordbuf[wordp++]=*p;
		}
		p++;
	}
	flush_word(0);
}


void renderer_reset(void)
{
	int boldspace, boldrow;
	int i;
	
	lastnl = 1;
	wordp = 0;
	pos_x = 0;
	
	for(i=0;i<3;i++)
	{
		spacesizes[i] = gdk_char_width(html_font[i][0], ' ');
		rowsizes[i] = gdk_char_height(html_font[i][0], 'X');

		boldspace = gdk_char_width(html_font[i][1], ' ');
		boldrow = gdk_char_height(html_font[i][1], 'X');
	
		/*
		 *	Allow for both fonts
		 */

		if(boldrow > rowsizes[i])
			rowsizes[i] = boldrow;
		if(boldspace > spacesizes[i])
			spacesizes[i] = boldspace;
	
		rowsizes[i]+=3;
	}		

	pos_y = rowsizes[0];

	*titlebuf=0;
}

void renderer_begin(struct token_stack *t)
{
	if(pixmap)
		gdk_draw_rectangle(pixmap,
			html->style->white_gc,
			TRUE,
			0, 0,
			html->allocation.width,
			html->allocation.height);
	link_free();
}


void render_newline(void)
{
	flush_word(0);
	pos_y += rowsize;
	lastnl=1;
	pos_x=0;
	rowsize=rowsizes[0];
}

void render_newline_p(void)
{
	flush_word(0);
	if(lastnl==0)
		render_newline();
}

void renderer_complete(struct token_stack *t)
{
	render_newline_p();
	GTK_ADJUSTMENT(scrollpos)->upper = (gfloat) pos_y;
	GTK_ADJUSTMENT(scrollpos)->step_increment = (gfloat) rowsize;
	GTK_ADJUSTMENT(scrollpos)->page_size = (gfloat)pixheight;
	if(pixheight > 2*rowsize)
		GTK_ADJUSTMENT(scrollpos)->page_increment = (gfloat)(pixheight-rowsize);
	else
		GTK_ADJUSTMENT(scrollpos)->page_increment = (gfloat)pixheight;
	gtk_adjustment_changed(GTK_ADJUSTMENT(scrollpos));
}



static gint resize_html(GtkWidget *w, GdkEventConfigure *event)
{
	static int n = 0;
	static int oldh= -1, oldw = -1;

	if(oldh != w->allocation.height || oldw != w->allocation.width)
	{
		if(!pixmap_dirty)
			pixmap_dirty = 1;
		
		if(pixmap)
			gdk_pixmap_unref(pixmap);
		pixmap = gdk_pixmap_new(w->window,
					w->allocation.width,
					w->allocation.height,
					-1);
		gdk_draw_rectangle(pixmap,
			w->style->white_gc,
			TRUE,
			0, 0,
			w->allocation.width,
			w->allocation.height);
		scrollbottom = scrolltop + w->allocation.height;
		pixwidth = w->allocation.width - 16;
		pixheight = w->allocation.height;
#if 0	
		render_reset(&gtk_rt);
		parser(pagebuf);
		parse_complete();
#endif
		oldh = w->allocation.height;
		oldw = w->allocation.width;
	}
	return TRUE;
}

static gint draw_html(GtkWidget *w, GdkEventExpose *event)
{
	if(pixmap_dirty == 1)
	{
		pixmap_dirty = 2;
		render_reset(&gtk_rt);
		parser(pagebuf);
		parse_complete();
		pixmap_dirty = 0;
	}
	
	if(pixmap)
		gdk_draw_pixmap(w->window,
			w->style->fg_gc[GTK_WIDGET_STATE(w)],
			pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
	return FALSE;
}

void mark_html_dirty(void)
{
	GdkRectangle update_rect;
	
	if(!pixmap_dirty)
	{
		pixmap_dirty = 1;

		update_rect.x = 0;
		update_rect.y = 0;
		update_rect.height = pixheight;
		update_rect.width = pixwidth + 16;

		gtk_widget_draw(html, &update_rect);
	}
}

void reset_scrollbar(void)
{
	gtk_adjustment_set_value(GTK_ADJUSTMENT(scrollpos),0.0);
}

static gint scroll_html(GtkWidget *w, gpointer unused)
{
	static int lastscroll;
	
	scrolltop = (int)GTK_ADJUSTMENT(scrollpos)->value;
	scrollbottom = scrolltop + pixheight;
	
	if(scrolltop != lastscroll)
	{
		lastscroll = scrolltop;
		mark_html_dirty();
	}
}

static gint click_html(GtkWidget *w, GdkEventButton *b)
{
	if(b->button == 1)
	{
		struct link_url *l = link_hit(b->x, b->y);
		if(l!=NULL)
		{
			printf("Go to %s\n", l->data);
			set_base_url(l->data);
			load_url();
		}
	}
}

int main(int argc, char *argv[])
{
	int i;
		/* FIXME */
	int fd=open(argv[1]?argv[1]:"foo.html", O_RDONLY);
	int l = read(fd, buf, 16384);
	close(fd);
	buf[l]=0;	

	
	gtk_init(&argc, &argv);
	app = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect(GTK_OBJECT(app), "destroy", 
		GTK_SIGNAL_FUNC(gtk_exit), NULL);
	gtk_window_set_title(GTK_WINDOW(app), "GWeb");
	gtk_container_set_border_width(GTK_CONTAINER(app), 0);
	
	draw = gtk_hbox_new(FALSE,0);
	gtk_container_set_border_width(GTK_CONTAINER(draw), 2);

	gtk_container_add(GTK_CONTAINER(app), draw);

	for(i=0;i<3;i++)
	{
		char fontbuf[128];
		sprintf(fontbuf, MAIN_FONT, "medium", 12+2*i);
		html_font[i][0] = gdk_font_load(fontbuf);
		if(html_font[i][0] == NULL)
		{
			fprintf(stderr, "Font not found!\n");
			exit(1);
		}
		gdk_font_ref(html_font[i][0]);
		sprintf(fontbuf, MAIN_FONT, "bold", 12+2*i);
		html_font[i][1] = gdk_font_load(fontbuf);
		if(html_font[i][1] == NULL)
		{
			fprintf(stderr, "Font not found!\n");
			exit(1);
		}
		gdk_font_ref(html_font[i][1]);
	}

	gdk_color_parse("#0000FF", &blue);
	gdk_colormap_alloc_color(gdk_colormap_get_system(), &blue, FALSE, TRUE);

	html = gtk_drawing_area_new();
	gtk_box_pack_start(GTK_BOX(draw), html, TRUE, TRUE, 0);
	
	scrollpos = gtk_adjustment_new(0.0,0.0,0.0,0.0,0.0,0.0);
	scroller = gtk_vscrollbar_new(GTK_ADJUSTMENT(scrollpos));
	gtk_box_pack_start(GTK_BOX(draw), scroller, FALSE, FALSE, 0);
	
	gtk_signal_connect(GTK_OBJECT(html), "expose_event", (GtkSignalFunc)draw_html, NULL);
	gtk_signal_connect(GTK_OBJECT(html), "configure_event", (GtkSignalFunc)resize_html, NULL);
	gtk_signal_connect(GTK_OBJECT(html), "button_press_event", (GtkSignalFunc)click_html, NULL);

	gtk_widget_set_events(html, GDK_EXPOSURE_MASK|GDK_BUTTON_PRESS_MASK);
		
	gtk_signal_connect(GTK_OBJECT(scrollpos), "value-changed", (GtkSignalFunc)scroll_html, NULL);
	
	gtk_widget_show(html);
	gtk_widget_show(scroller);
	gtk_widget_show(draw);
	gtk_widget_show(app);
	gtk_main();
	return 0;
}
